import logging
import time
import math

import grpc

DEFAULT_OPTIONS = {
    "min_delay": 1,
    "max_delay": 50,
    "delay_factor": math.exp(1),
    "max_retries": 5
}
RETRY_CODES = [1, 2, 4, 10, 13, 14, 15]


class AuthenticatedService(object):
    """Initialization class for gRPC services.

    """
    def __init__(self, auth_token_func, stub, verbose=False, options=None):
        self.stub = stub
        self.auth_token_func = auth_token_func
        self.verbose = verbose
        if options:
            self.min_delay, self.max_delay, self.delay_factor, self.max_retries = self.parse_options(options)
        else:
            self.min_delay, self.max_retries = 0, 0
        for method_name in dir(stub):
            method = getattr(stub, method_name)
            if not method_name.startswith('__') and callable(method):
                setattr(self, method_name, self.__decorated(method_name, method))

    def __decorated(self, method_name, method):
        def inner(param):
            auth_token = self.auth_token_func()
            if self.verbose:
                logging.info('Calling %s with token=`%s`' % (method_name, auth_token))
            if auth_token is not None:
                metadata = (('x-auth-ticket', auth_token),)
            else:
                metadata = None
            tries = 0
            delay = self.min_delay
            while 1:
                try:
                    return method(param, metadata=metadata)
                except grpc.RpcError as e:
                    if e._state.code.value[0] not in RETRY_CODES:
                        logging.error("Failed request to server, with error: " + e.details(), e.debug_error_string())
                        raise e
                    if self.max_retries > tries:
                        time.sleep(delay)
                        tries += 1
                        delay = min(delay * self.delay_factor, self.max_delay)
                        logging.error(str(e))
                        continue
                    logging.error("Max retries requests to server, with error: " + e.details(), e.debug_error_string())
                    raise e
        return inner

    @staticmethod
    def parse_options(options):
        for option, value in DEFAULT_OPTIONS.items():
            if option not in options:
                options[option] = value
        return options["min_delay"], options["max_delay"], options["delay_factor"], options["max_retries"]



import random
from typing import List

from google.protobuf import wrappers_pb2

from dialog_bot_sdk.entities.Avatar import Avatar
from dialog_bot_sdk.entities.Group import Group
from dialog_bot_sdk.entities.Permissions import Permissions, GroupPermission
from dialog_bot_sdk.entities.User import User
from dialog_bot_sdk.entities.Peer import Peer, PeerType
from dialog_bot_sdk.utils import async_dec, AsyncTask, get_peer
from .service import ManagedService
from dialog_api import search_pb2, groups_pb2, peers_pb2, sequence_and_updates_pb2


class Groups(ManagedService):
    """Class for handling groups

    """
    @async_dec()
    def create_public_group(self, title: str, short_name: str) -> Group:
        """Create public group

        :param title: title of group
        :param short_name: group name
        :return: Group
        """
        request = groups_pb2.RequestCreateGroup(
            title=title,
            username=wrappers_pb2.StringValue(value=short_name),
            group_type=groups_pb2.GROUPTYPE_GROUP
        )
        return self.__create_group(request)

    @async_dec()
    def create_private_group(self, title: str) -> Group:
        """Create private group

        :param title: title of group
        :return: Group
        """
        request = groups_pb2.RequestCreateGroup(
            title=title,
            group_type=groups_pb2.GROUPTYPE_GROUP
        )
        return self.__create_group(request)

    @async_dec()
    def create_public_channel(self, title: str, short_name: str) -> Group:
        """Create public channel

        :param title: title of group
        :param short_name: group name
        :return: Group
        """

        request = groups_pb2.RequestCreateGroup(
            title=title,
            username=wrappers_pb2.StringValue(value=short_name),
            group_type=groups_pb2.GROUPTYPE_CHANNEL
        )
        return self.__create_group(request)

    @async_dec()
    def create_private_channel(self, title: str) -> Group:
        """Create private channel

        :param title: title of group
        :return: Group
        """

        request = groups_pb2.RequestCreateGroup(
            title=title,
            group_type=groups_pb2.GROUPTYPE_CHANNEL
        )
        return self.__create_group(request)

    @async_dec()
    def find_group_by_short_name(self, short_name: str) -> Group or None:
        """Find a Group by short_name

        :param short_name: short_name of group
        :return: Group or None if could not find
        """
        request = search_pb2.RequestPeerSearch(
            query=[
                search_pb2.SearchCondition(
                    searchPeerTypeCondition=search_pb2.SearchPeerTypeCondition(
                        peer_type=search_pb2.SEARCHPEERTYPE_GROUPS
                    )
                ),
                search_pb2.SearchCondition(
                    searchPieceText=search_pb2.SearchPieceText(query=short_name)
                )
            ]
        )
        response = self.internal.search.PeerSearch(request).search_results
        for result in response:
            if result.peer.type == PeerType.PEERTYPE_GROUP and hasattr(result, 'shortname') and \
                    result.shortname.value == short_name:
                return self.find_group_by_id(result.peer.id).wait()

    @async_dec()
    def find_group_by_id(self, group_id: int) -> Group or None:
        """Find and return Group by id

        :param group_id: group's
        :return: Group or None if could not find
        """
        out_peer = self.__get_out_peer(Peer(group_id, PeerType.PEERTYPE_GROUP))
        if out_peer is None:
            return None

        request = sequence_and_updates_pb2.RequestGetReferencedEntitites(
            groups=[out_peer]
        )
        result = self.internal.updates.GetReferencedEntitites(request).groups
        self.manager.add_out_peer(peers_pb2.OutPeer(id=result[0].id, type=PeerType.PEERTYPE_GROUP,
                                                    access_hash=result[0].access_hash))
        return Group.from_api(result[0])

    @async_dec()
    def load_members(self, group_peer: Peer or AsyncTask, limit: int = 0) -> List[User] or None:
        """Load Group members by peer

        :param group_peer: Peer or AsyncTask (in which located Group)
        :param limit: count members
        :return: list of User's
        """
        group_peer = get_peer(group_peer)
        out_peer = self.__get_out_peer(group_peer)
        if out_peer is None:
            return None

        request = groups_pb2.RequestLoadMembers(
            group=out_peer,
            limit=limit
        )
        members = self.internal.groups.LoadMembers(request).members
        request = sequence_and_updates_pb2.RequestGetReferencedEntitites(
            group_members=[
                sequence_and_updates_pb2.GroupMembersSubset(
                    group_peer=out_peer,
                    member_ids=[member.uid for member in members]
                )
            ]
        )
        return [User.from_api(x) for x in self.internal.updates.GetReferencedEntitites(request).users]

    @async_dec()
    def kick_user(self, group_peer: Peer or AsyncTask, user_peer: Peer or AsyncTask) -> None:
        """Kick user from Group

        :param group_peer: Peer or AsyncTask (in which located Group)
        :param user_peer: Peer or AsyncTask (in which located User)
        :return: None
        """
        group_peer, user_peer = get_peer(group_peer), get_peer(user_peer)
        group_out_peer, user_out_peer = self.__get_out_peer(group_peer), self.__get_out_peer(user_peer)

        request = groups_pb2.RequestKickUser(
            group_peer=group_out_peer,
            user=user_out_peer,
            rid=random.randint(0, 100000000),
        )
        self.internal.groups.KickUser(request)

    @async_dec()
    def invite_user(self, group_peer: Peer or AsyncTask, user_peer: Peer or AsyncTask) -> None:
        """Invite user in Group

        :param group_peer: Peer or AsyncTask (in which located Group)
        :param user_peer: Peer or AsyncTask (in which located User)
        :return: None
        """
        group_peer, user_peer = get_peer(group_peer), get_peer(user_peer)
        group_out_peer, user_out_peer = self.__get_out_peer(group_peer), self.__get_out_peer(user_peer)

        request = groups_pb2.RequestInviteUser(
            group_peer=group_out_peer,
            user=user_out_peer,
            rid=random.randint(0, 100000000),
        )
        self.internal.groups.InviteUser(request)

    @async_dec()
    def set_default_group_permissions(self, group_peer: Peer or AsyncTask,
                                      add_permissions: List[GroupPermission] = None,
                                      del_permissions: List[GroupPermission] = None) -> None:
        """add/del default group permissions

        :param group_peer: Peer or AsyncTask (in which located Group)
        :param add_permissions: list of permissions to add
        :param del_permissions: list of permissions to delete
        :return: None
        """
        group_peer = get_peer(group_peer)
        if del_permissions is None:
            del_permissions = []
        if add_permissions is None:
            add_permissions = []

        group_out_peer = self.__get_out_peer(group_peer)
        add_request = groups_pb2.RequestEditGroupBasePermissions(
            group_peer=group_out_peer,
            random_id=random.randint(0, 100000000),
            granted_permissions=[x for x in add_permissions]
        )
        del_request = groups_pb2.RequestEditGroupBasePermissions(
            group_peer=group_out_peer,
            random_id=random.randint(0, 100000000),
            revoked_permissions=[x for x in del_permissions]
        )

        self.internal.groups.EditGroupBasePermissions(add_request)
        self.internal.groups.EditGroupBasePermissions(del_request)

    @async_dec()
    def set_member_permissions(self, group_peer: Peer or AsyncTask, user_peer: Peer or AsyncTask,
                               add_permissions: List[GroupPermission] = None,
                               del_permissions: List[GroupPermission] = None) -> None:
        """add/del group's member permissions

        :param group_peer: Peer or AsyncTask (in which located Group)
        :param user_peer: Peer or AsyncTask (in which located User)
        :param add_permissions: list of permissions to add
        :param del_permissions: list of permissions to delete
        :return: None
        """
        group_peer, user_peer = get_peer(group_peer), get_peer(user_peer)
        if del_permissions is None:
            del_permissions = []
        if add_permissions is None:
            add_permissions = []
        group_out_peer, user_out_peer = self.__get_out_peer(group_peer), self.__get_out_peer(user_peer)
        add_request = groups_pb2.RequestEditMemberPermissions(
            group_peer=group_out_peer,
            user_peer=user_out_peer,
            granted_permissions=[x for x in add_permissions]
        )
        del_request = groups_pb2.RequestEditMemberPermissions(
            group_peer=group_out_peer,
            user_peer=user_out_peer,
            revoked_permissions=[x for x in del_permissions]
        )

        self.internal.groups.EditMemberPermissions(add_request)
        self.internal.groups.EditMemberPermissions(del_request)

    @async_dec()
    def get_group_member_permissions(self, group_peer: Peer or AsyncTask, user_peers: List[Peer or AsyncTask]) \
            -> List[Permissions]:
        """return group member's permissions

        :param group_peer: Peer or AsyncTask (in which located Group)
        :param user_peers: Peer or AsyncTask (in which located User)
        :return: group member's permissions
        """
        group_peer, user_peers = get_peer(group_peer), [get_peer(x) for x in user_peers]
        request = groups_pb2.RequestGetGroupMemberPermissions(
            group_id=group_peer.id,
            user_ids=[peer.id for peer in user_peers]
        )
        return [Permissions.from_api(x) for x in self.internal.groups.GetGroupMemberPermissions(request).permissions]

    @async_dec()
    def edit_group_title(self, group_peer: Peer or AsyncTask, title: str) -> None:
        """change group's title

        :param group_peer: Peer or AsyncTask (in which located Group)
        :param title: new title
        :return: None
        """
        group_peer = get_peer(group_peer)
        out_peer = self.__get_out_peer(group_peer)
        request = groups_pb2.RequestEditGroupTitle(
            group_peer=out_peer,
            rid=random.randint(0, 100000000),
            title=title
        )
        self.internal.groups.EditGroupTitle(request)

    @async_dec()
    def edit_avatar(self, group_peer: Peer or AsyncTask, file: str) -> Avatar or None:
        """change group's avatar

        :param group_peer: Peer or AsyncTask (in which located Group)
        :param file: file path
        :return: Avatar
        """
        group_peer = get_peer(group_peer)
        out_peer = self.__get_out_peer(group_peer)
        location = self.internal.uploading.upload_file(file).wait()
        if location is None:
            return None
        request = groups_pb2.RequestEditGroupAvatar(
            group_peer=out_peer,
            rid=random.randint(0, 100000000),
            file_location=location.to_api()
        )
        return Avatar.from_api(self.internal.groups.EditGroupAvatar(request).avatar)

    @async_dec()
    def remove_group_avatar(self, group_peer: Peer or AsyncTask) -> None:
        """deleted group's avatar

        :param group_peer: Peer or AsyncTask (in which located User)
        :return: None
        """
        group_peer = get_peer(group_peer)
        out_peer = self.__get_out_peer(group_peer)
        request = groups_pb2.RequestRemoveGroupAvatar(
            group_peer=out_peer,
            rid=random.randint(0, 100000000),
        )
        self.internal.groups.RemoveGroupAvatar(request)

    @async_dec()
    def edit_group_about(self, group_peer: Peer or AsyncTask, about: str) -> None:
        """change group's "about"

        :param group_peer: Peer or AsyncTask (in which located User)
        :param about: about text
        :return: None
        """
        group_peer = get_peer(group_peer)
        out_peer = self.__get_out_peer(group_peer)
        request = groups_pb2.RequestEditGroupAbout(
            group_peer=out_peer,
            rid=random.randint(0, 100000000),
            about=wrappers_pb2.StringValue(value=about)
        )
        self.internal.groups.EditGroupAbout(request)

    @async_dec()
    def leave_group(self, group_peer: Peer or AsyncTask) -> None:
        """leave from group

        :param group_peer: Peer or AsyncTask (in which located Group)
        :return: None
        """
        group_peer = get_peer(group_peer)
        out_peer = self.__get_out_peer(group_peer)
        request = groups_pb2.RequestLeaveGroup(
            group_peer=out_peer,
            rid=random.randint(0, 100000000),
        )
        self.internal.groups.LeaveGroup(request)

    @async_dec()
    def make_user_admin(self, group_peer: Peer or AsyncTask, user_peer: Peer or AsyncTask,
                        permissions: List[GroupPermission]) -> None:
        """Set new user's permissions (old permissions will be revoke)

        :param group_peer: Peer or AsyncTask (in which located Group)
        :param user_peer: Peer or AsyncTask (in which located User)
        :param permissions: permissions list (for admin)
        :return: None
        """
        group_peer, user_peer = get_peer(group_peer), get_peer(user_peer)
        group_out_peer, user_out_peer = self.__get_out_peer(group_peer), self.__get_out_peer(user_peer)
        request = groups_pb2.RequestMakeUserAdmin(
            group_peer=group_out_peer,
            user_peer=user_out_peer,
            permissions=permissions
        )
        self.internal.groups.MakeUserAdmin(request)

    @async_dec()
    def transfer_ownership(self, group_peer: Peer or AsyncTask, user_peer: Peer or AsyncTask) -> None:
        """change group's owner to user

        :param group_peer: Peer or AsyncTask (in which located Group)
        :param user_peer: Peer or AsyncTask (in which located User)
        :return: None
        """
        group_peer, user_peer = get_peer(group_peer), get_peer(user_peer)
        out_peer = self.__get_out_peer(group_peer)
        request = groups_pb2.RequestTransferOwnership(
            group_peer=out_peer,
            new_owner=user_peer.id
        )
        self.internal.groups.TransferOwnership(request)

    @async_dec()
    def get_group_invite_url(self, group_peer: Peer or AsyncTask) -> str:
        """return group's invite url

        :param group_peer: Peer or AsyncTask (in which located Group)
        :return: invite url
        """
        group_peer = get_peer(group_peer)
        out_peer = self.__get_out_peer(group_peer)
        request = groups_pb2.RequestGetGroupInviteUrl(
            group_peer=out_peer
        )
        return self.internal.groups.GetGroupInviteUrl(request).url

    @async_dec()
    def get_group_invite_url_base(self) -> str:
        """return group's invite url without token/short_name (example https://domain/@)

        :return: invite url (string)
        """
        request = groups_pb2.RequestGetGroupInviteUrlBase()
        return self.internal.groups.GetGroupInviteUrlBase(request).url

    @async_dec()
    def revoke_invite_url(self, group_peer: Peer or AsyncTask) -> str:
        """revoke current invite url and return new group's invite url

        :return: invite url
        """
        group_peer = get_peer(group_peer)
        out_peer = self.__get_out_peer(group_peer)
        request = groups_pb2.RequestRevokeInviteUrl(
            group_peer=out_peer
        )
        return self.internal.groups.RevokeInviteUrl(request).url

    @async_dec()
    def join_group(self, token_or_url: str) -> Group:
        """join to group by token or invite url (used for private groups)

        :param token_or_url: group's token or invite url
        :return: Group
        """
        request = groups_pb2.RequestJoinGroup(
            token=token_or_url
        )
        response = self.internal.groups.JoinGroup(request)
        self.manager.add_out_peer(peers_pb2.OutPeer(id=response.group.id, access_hash=response.group.access_hash,
                                                    type=PeerType.PEERTYPE_GROUP))
        return Group.from_api(response.group)

    @async_dec()
    def join_group_by_peer(self, group_peer: Peer or AsyncTask) -> None:
        """join to group by group's peer (used for public groups)

        :param group_peer: Peer or AsyncTask (in which located Group)
        :return: None
        """
        group_peer = get_peer(group_peer)
        out_peer = self.__get_out_peer(group_peer)
        request = groups_pb2.RequestJoinGroupByPeer(
            peer=out_peer
        )
        self.internal.groups.JoinGroupByPeer(request)

    def __get_out_peer(self, peer: peers_pb2.Peer) -> peers_pb2.GroupOutPeer or peers_pb2.UserOutPeer or None:
        out_peer = self.manager.get_out_peer(peer)
        if out_peer is None:
            return None
        if peer.type == PeerType.PEERTYPE_GROUP:
            return peers_pb2.GroupOutPeer(group_id=out_peer.id, access_hash=out_peer.access_hash)
        elif peer.type == PeerType.PEERTYPE_PRIVATE:
            return peers_pb2.UserOutPeer(uid=out_peer.id, access_hash=out_peer.access_hash)

    def __create_group(self, request: groups_pb2.RequestCreateGroup) -> Group:
        group = self.internal.groups.CreateGroup(request).group
        self.manager.add_out_peer(
            peers_pb2.OutPeer(id=group.id, access_hash=group.access_hash, type=PeerType.PEERTYPE_GROUP))
        return Group.from_api(group)



import os
import mimetypes

from dialog_api import messaging_pb2
from dialog_bot_sdk.utils import get_image_w_h, get_image_thumb_bytes


def get_document_content(file, location):
    """Prepares DocumentMessage object that contains metadata for file sending request.

    :param file: path to file
    :param location: FileLocation object
    :return: DocumentMessage object
    """

    content = messaging_pb2.DocumentMessage()

    content.file_id = location.file_id
    content.access_hash = location.access_hash
    content.name = os.path.basename(file)
    content.file_size = os.path.getsize(file)

    return content


def get_image_content(file, location):
    """Same as get_document_content for image sending.

    :param file: path to image
    :param location: FileLocation object
    :return: DocumentMessage object
    """

    content = get_document_content(file, location)

    content.mime_type = mimetypes.guess_type(file)[0]
    content.thumb.w, content.thumb.h, content.thumb.thumb = get_image_thumb_bytes(file)

    ext = messaging_pb2.DocumentEx()

    photo = messaging_pb2.DocumentExPhoto()
    photo.w, photo.h = get_image_w_h(file)
    ext.photo.CopyFrom(photo)
    content.ext.CopyFrom(ext)

    return content



from dialog_bot_sdk.entities.Peer import Peer, PeerType
from dialog_bot_sdk.exceptions.exceptions import UnknownPeerError
from dialog_bot_sdk.internal.bot import InternalBot

from dialog_api import peers_pb2, messaging_pb2


class EntityManager(object):
    """Entity manager class.

    Add OutPeer's to dict when bot receive a message, to avoid going to the server once again.
    """
    def __init__(self, internal: InternalBot) -> None:
        self.internal = internal
        self.peer_to_out_peer = {}

    def get_out_peer(self, peer: Peer or peers_pb2.Peer) -> peers_pb2.OutPeer or None:
        """Returns OutPeer for given Peer.

        :param peer: Peer
        :return: OutPeer
        """
        if isinstance(peer, Peer):
            peer = peer.to_api()
        if not isinstance(peer, peers_pb2.Peer):
            raise RuntimeError("Invalid input data. Expects {}, got {}.".format(Peer.__class__, type(peer)))
        result = self.peer_to_out_peer.get((peer.type, peer.id))
        if result is None:
            req = messaging_pb2.RequestLoadDialogs(
                min_date=0,
                limit=1,
                peers_to_load=[peer],
            )
            result = self.internal.messaging.LoadDialogs(req)
            for user in result.user_peers:
                self.__adopt_peer(user)
            for group in result.group_peers:
                self.__adopt_peer(group)
            return self.peer_to_out_peer.get((peer.type, peer.id))
        return result

    def __adopt_peer(self, peer: peers_pb2.UserOutPeer or peers_pb2.GroupOutPeer) -> None:
        """Find outpeer for given peer and store it in internal peers_to_outpeers dict.

        :param peer: Peer object (UserOutPeer or GroupOutPeer)
        """
        if isinstance(peer, peers_pb2.UserOutPeer):
            out_peer = peers_pb2.OutPeer(id=peer.uid, access_hash=peer.access_hash, type=PeerType.PEERTYPE_PRIVATE)
        elif isinstance(peer, peers_pb2.GroupOutPeer):
            out_peer = peers_pb2.OutPeer(id=peer.group_id, access_hash=peer.access_hash, type=PeerType.PEERTYPE_GROUP)
        else:
            raise UnknownPeerError("Unknown PeerType.")
        self.peer_to_out_peer[(out_peer.type, out_peer.id)] = out_peer

    def add_out_peer(self, out_peer: peers_pb2.OutPeer) -> None:
        """Adding OutPeers to peer_to_out_peer

        :param out_peer: OutPeer
        :return: None
        """
        self.peer_to_out_peer[(out_peer.type, out_peer.id)] = out_peer



import io
import logging
import mimetypes
import sys
import time
from typing import List
from concurrent.futures import ThreadPoolExecutor

import six
from PIL import Image
from dialog_api import peers_pb2, media_and_files_pb2, messaging_pb2

from dialog_bot_sdk.entities.Peer import Peer, PeerType
from dialog_bot_sdk.entities.UUID import UUID
from dialog_bot_sdk.exceptions.exceptions import UnknownPeerError


POOL = ThreadPoolExecutor(max_workers=10)


def get_image_w_h(file):
    """Returns image's width and height

    :param file: path to image
    :return: (width, height) tuple
    """

    im = Image.open(file)
    return im.size


def get_image_thumb_bytes(file):
    """Returns image's thumbnail in form of byte array

    :param file: path to image
    :return: size and byte array
    """

    size = 90, 90

    im = Image.open(file)
    im = im.convert('RGB')
    im.thumbnail(size)
    stream = io.BytesIO()
    im.save(stream, "JPEG")

    return size[0], size[1], stream.getvalue()


def get_image_location(bot, file, width=100, height=100):
    """Return obj ImageLocation

    :param bot: DialogBot
    :param file: image's file
    :param width: image's width
    :param height: image's height
    :return: ImageLocation
    """
    if not is_image(file):
        raise IOError('File is not an image.')
    location = bot.internal.uploading.upload_file(file)
    return media_and_files_pb2.ImageLocation(file_location=location, width=width, height=height)


def get_image(bot, file, width=100, height=100):
    """Return MessageMedia with ImageMedia for messaging.send_media

    :param bot: DialogBot
    :param file: image's file
    :param width: subj
    :param height: subj
    :return: MessageMedia obj
    """
    image_location = get_image_location(bot, file, width, height)
    return messaging_pb2.MessageMedia(image=messaging_pb2.ImageMedia(image=image_location))


def is_image(file):
    mime = mimetypes.guess_type(file)[0]
    return mime in ['image/jpeg', 'image/png', 'image/gif', 'image/bmp']


def read_file_in_chunks(file, chunk_size=1024*1024):
    """File chunks generator

    :param file: path to file
    :param chunk_size: size of a chunk
    :return: generator object that iterates over file chunks
    """
    file_object = open(file, 'rb')
    while True:
        data = file_object.read(chunk_size)
        if not data:
            file_object.close()
            break
        yield data


class AsyncTask:
    def __init__(self, target, *args, **kwargs):
        self.target = target
        self.args = args
        self.kwargs = kwargs

        self.done = False
        POOL.submit(self._run)

    def _run(self):
        try:
            self.result = self.target(*self.args, **self.kwargs)
        except:
            self.result = sys.exc_info()
            logging.error(self.result[0], self.result[1], self.result[2])
            six.reraise(self.result[0], self.result[1], self.result[2])
        self.done = True

    def wait(self):
        while not self.done:
            time.sleep(0.01)
        if isinstance(self.result, BaseException):
            logging.error(self.result[0], self.result[1], self.result[2])
            six.reraise(self.result[0], self.result[1], self.result[2])
        else:
            logging.debug("target: {}\nargs: {}\nkwargs: {}\nresult: {}\n"
                          .format(self.target, self.args, self.kwargs, self.result))
            return self.result


def async_dec():
    def decorator(fn):
        def wrapper(*args, **kwargs):
            return AsyncTask(fn, *args, **kwargs)

        return wrapper

    return decorator


def get_peer(peer: Peer or AsyncTask) -> peers_pb2.Peer or None:
    if isinstance(peer, AsyncTask):
        from_task = peer.wait()
        if not hasattr(from_task, 'peer'):
            raise AttributeError("{} has not attribute 'peer'".format(type(from_task)))
        peer = from_task.peer
    if not isinstance(peer, Peer):
        raise AttributeError("peer must be {}, got {}.".format(Peer, type(peer)))
    if peer.type in [PeerType.PEERTYPE_PRIVATE, PeerType.PEERTYPE_GROUP]:
        return peers_pb2.Peer(id=peer.id, type=peer.type)
    else:
        raise UnknownPeerError("Unknown PeerType.")


def get_uuids(uuids: List[UUID or AsyncTask]):
    for i in range(len(uuids)):
        uuid = uuids[i]
        if isinstance(uuid, AsyncTask):
            uuid = uuid.wait()
        if isinstance(uuid, UUID):
            uuids[i] = uuid.to_api()
        else:
            raise AttributeError("uuid must be {}, got {}.".format(UUID.__class__, type(uuid)))
    return uuids
